package com.kpro.algorithm;

import java.util.ArrayList;

import com.kpro.dataobjects.Action;
import com.kpro.dataobjects.PolicyObject;

/**
 * A very simple conclusion class. Result is based on the closest objects only, as determined by the sum of inverse distances
 * of the accepted versus rejected policies. Confidences is the ratio of sum inverse distances of the chosen decision, versus the sum
 * of all inverse distances.
 * 
 * @author ngerstle
 * @version 29.09.11.1
 */
public class ConclusionSimple extends ConclusionAlgorithm {
	
	public ConclusionSimple(DistanceMetric distanceMetric, String[] extraArgs) {
		super(distanceMetric, extraArgs);
	}

	/**
	 * Makes a decision on the reduced set.
	 * This class creates two lists, one for accepted policies and one for rejected. Assuming there are policies in both 
	 * (easy decision otherwise), whether the policy is accepted or not will depend on the difference between the sum of inverse
	 * distances of the list items (excluding zero-distances), with the smaller sum indicating the more relevant decision.
	 * 
	 * @author ngerstle
	 * 
	 * @param policyObject The object under consideration
	 * @param reducedNeighbors the reduced set of neighbors
	 * @return an ArrayList of {Action a, double Confidence) 
	 *
	 */
	@Override
	public Action conclude(PolicyObject policyObject, Iterable<PolicyObject> reducedNeighbors) {
		ArrayList<String> approveList = new ArrayList<String>();
		ArrayList<String> rejectList = new ArrayList<String>();
		
		double approveDistance = 0; //the sum of the inverse distance to all approved PolicyObjects
		double rejectDistance = 0; //the sum of the inverse distance to all rejected PolicyObjects

		for(PolicyObject po : reducedNeighbors) {
			if(po.getAction() != null) {
				double d = distanceMetric.getTotalDistance(policyObject, po); //just in case this is zero...
				
				if(po.getAction().getAccepted()) {
					approveList.add(po.getContextDomain());
					approveDistance += (d==0) ? (0):(1/d);
				} else {
					rejectList.add(po.getContextDomain());
					rejectDistance += (d==0) ? (0):(1/d);				
				}
			}
		}

//		System.err.print("approveList \t"+approveList+"\t\t confidence \t"+ (appdistance/(appdistance+rejdistance)));
//		System.err.print("rejectList \t"+rejectList+"\t\t confidence \t"+ (rejdistance/(appdistance+rejdistance)) );
//		System.err.print("\t\t appdistance: "+appdistance +"\trejdistance"+rejdistance);
		
		if(approveList.isEmpty() && rejectList.isEmpty()) {
			System.err.println("c=0");
			return new Action(false, approveList, 0, false);
		} else if(approveList.isEmpty()) {
			return new Action(false, rejectList, 1, false);
		} else if(rejectList.isEmpty()) {
			return new Action(true, approveList, 1, false);
		} else {
			if(approveDistance > rejectDistance) {
				return new Action(true, approveList, 
						(approveDistance/(approveDistance+rejectDistance)), false);
			} else {
				return new Action(false, rejectList, 
						(rejectDistance/(approveDistance+rejectDistance)), false);
			}
		}
	}
}